/*
  Jeremy Weatherford <xidus TA xidus TOD net>
  orbitclock.c

  No copyright -- steal this code!

  gcc -o orbitclock orbitclock.c -lX11 -L/usr/X11R6/lib

  v0.1 -- compiles on Irix, and is ANSIer C, thanks to jonh TA sgi TOD com
  v0.2 -- double-buffered, thanks to hugh4242 TA yahoo TOD com for pointer
          command line options for -geometry, -bg, -fg, -fg2, -rings, -date
	  rearrangement of rings thanks to lornix TA lornix TOD com
	  date option for including text in center
  v0.3 -- adapted to suit kernel scheduler benchmarks - Willy Tarreau
  v0.4 -- support SIGUSR1 to reset all counters
  v0.5 -- make windows unmanaged with -u - Bjrn Steinbrink
*/


#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <getopt.h>

#include <X11/Xlib.h>

#define min(x,y) (((x) < (y)) ? (x) : (y))
#define max(x,y) (((x) > (y)) ? (x) : (y))

/* default options */
int width=100, height=100, X=200, Y=200;
char *bg = "black", *bg2 = "green3", *fg = "white", *fg2 = "blue";

int HOUR=0;
int MIN =1;
int SEC =2;
int drawdate = 0;
char *datestr = "%m/%d";
int datelines = 0;
char *fontname = "fixed";

/* command-line options */
static struct option long_opts[] = {
  {"help",      no_argument, 0, 'h'},
  {"geometry",  required_argument, 0, 'g'},
  {"bg",        required_argument, 0, 'b'},
  {"fg",        required_argument, 0, '1'},
  {"fg2",       required_argument, 0, '2'},
  {"rings",     required_argument, 0, 'r'},
  {"date",      optional_argument, 0, 'd'},
  {"font",      required_argument, 0, 'f'},
  {"run",       required_argument, 0, 'R'},
  {"sleep",     required_argument, 0, 'S'},
  {0, 0, 0, 0}
};
const char *optstring = "uhg:b:1:2:r:d::f:R:S:x:y:B:";


void usage(void) {
  fprintf(stderr, "Usage: orbitclock [options]\n"
	  "\t-geometry WxH[+X+Y]     set window geometry\n"
	  "\t-bg <color name>        background color\n"
	  "\t-fg <color name>        foreground 1 color (tick)\n"
	  "\t-fg2 <color name>       foreground 2 color (block)\n"
	  "\t-rings <hms|smh|etc.>   ordering of rings from inside out (default hms)\n"
	  "\t-font <font>            font to use for date (implies -date)\n"
	  "\t-date[=optional format] draw date in center, strftime format string\n"
	  "\t                        literal '\\n' causes line break\n"
	  "\t                        example: -date=%%m/%%d/%%Y\\n%%H:%%M:%%S\n"
	  "\t                        equals sign is REQUIRED, don't ask me why.\n"
	  "\t-R <run_time_us>        CPU burning time in microseconds per iteration.\n"
	  "\t-S <sleep_time_us>      sleeping time in microseconds per iteration.\n"
	  "\t-x <nb_processes_x>     number of processes per lines (will fork).\n"
	  "\t-y <nb_processes_y>     number of processes per column (will fork).\n"
	  "\t-B <border size>        space to skip between multiple windows.\n"
	  "\t-u                      make windows unmanaged (recommended).\n"
	  );
}



/* globals, yuck */
Display *dpy = NULL;
Window win;
GC gc, copygc;
XColor black, orange, blue, green;
XFontStruct *font;
Pixmap buffer;

long long rounds = 0;
unsigned int runtime = 0;
unsigned int sleeptime = 10000;
unsigned int nbproc_x = 1, nbproc_y = 1, bdrsize = 0, unmanaged = 0;

void draw(void) {
  const int circle = 360*64;
  const int center = min(width,height)/2;// - 10;
  int gap = -3*64*((200-min(center,150))/50);
  int offset = 90 * 64;
  int slice = center/3 - 1;
  int rank;
  int i;
  struct tm tm;
  int counts[3];

  counts[HOUR] = (rounds / 3600) % 60;
  counts[MIN] = (rounds / 60) % 60;
  counts[SEC] = rounds % 60;
  rounds++;

  // clear to background
  XSetForeground(dpy, gc, black.pixel);
  XFillRectangle(dpy, buffer, gc, 0, 0, width, height);

  /* XFillArc doesn't let you give an inside radius, so we draw the
     outside circles first, always clearing the center to background */
  for (rank=2; rank >= 0; rank--) {
    int outer = center - ((2-rank)*slice);
    int inner = center-(slice-3) - ((2-rank)*slice);
    int count = (rank==HOUR) ? 12 : 60;
    int span = -circle / count;
    int group = (rank == HOUR) ? 3 : 5;
    XSetForeground(dpy, gc, orange.pixel);
    
    for(i=0; i<count; i++) {
      int start = span*i+gap/2;
      int width;
      if (i >= counts[rank]) {
	width=span-gap/3;
	XSetForeground(dpy, gc, green.pixel);
      } else if (i+group <= counts[rank]) {
	width = span*group-gap/3;
	XSetForeground(dpy, gc, blue.pixel);
	i += group-1;
      } else {
	width = span-gap/3;
	XSetForeground(dpy, gc, orange.pixel);
      }
      XFillArc(dpy, buffer, gc, center-outer, center-outer,
	       outer*2, outer*2, offset+start, width);
    }
    XSetForeground(dpy, gc, black.pixel);
    XFillArc(dpy, buffer, gc, center-inner, center-inner,
	     inner*2, inner*2, 0, circle);
  }
  
  if (drawdate) {
    int h, w, y, hpad;
    char buf[512];
    int len;
    char *p, *q;

    strftime(buf, sizeof(buf), datestr, &tm);

    // font height
    h = font->max_bounds.ascent + font->max_bounds.descent;
    hpad = h + 2;

    p = buf;
    y = center - (hpad*datelines)/2 - h/2;
    // for each line
    do {
      q = strchr(p, '\n');
      if (q) len = q-p; else len = strlen(p);
      
      w = XTextWidth(font, p, len);

      XSetForeground(dpy, gc, orange.pixel);
      XDrawString(dpy, buffer, gc, center-w/2, y+font->ascent, p, len);

      y += hpad;

      p = q; if (p) p++;
    } while (p && q);
  }

  // update window from buffer
  XSetTile(dpy, copygc, buffer);
  XFillRectangle(dpy, win, copygc, 0, 0, width, height);
}

void expose(XExposeEvent *ev) {
  if (ev->count > 0) return;
  draw();
}

int errhandler(Display *dpy, XErrorEvent *ev) {
  char buf[512];

  XGetErrorText(dpy, ev->error_code, buf, sizeof(buf));
  printf("XError %d, %s\n", ev->error_code, buf);
  exit(1);
  return 1;
}

void alloc_color(char *color, XColor *ret) {
  XColor dummy;
  Colormap cmap = DefaultColormap(dpy, DefaultScreen(dpy));

  if (!XAllocNamedColor(dpy, cmap, color, ret, &dummy)) {
    fprintf(stderr, "Failed to allocate color '%s'\n", color);
    exit(1);
  }
}

void sig_resync(int sig) {
    rounds = 0;
    signal(sig, sig_resync);
}

int main(int argc, char *argv[]) {
  Window root;
  XGCValues gc_setup;
  XSetWindowAttributes swa;
  int c, index, proc_x, proc_y, pid;
  int *pcount[] = {&HOUR, &MIN, &SEC};
  char *p, *q;

  while ((c = getopt_long_only(argc, argv, optstring, long_opts, &index)) != -1) {
    switch (c) {
    case 'h':
      usage();
      exit(0);
      break;

    case 'g':
      XParseGeometry(optarg, &X, &Y, &width, &height);
      break;

    case 'b':
      bg = strdup(optarg);
      break;

    case '1':
      fg = strdup(optarg);
      break;

    case '2':
      fg2 = strdup(optarg);
      break;

    case 'r':
      for (c=0; c<3; c++) {
	p = strchr(optarg, "hms"[c]);
	if (!p) { usage(); exit(1); }
	*pcount[c] = (int)(p-optarg);
      }
      break;

    case 'd':
      if (optarg != NULL) {
	datestr = strdup(optarg);

	datelines = 0;
	/* convert literal "\\n" to '\n' */
	for(p=datestr, q=datestr; *p; ) {
	  if (*p == '\\' && p[1] == 'n') {
	    *q++ = '\n';
	    datelines++;
	    p += 2;;
	  } else
	    *q++ = *p++;
	}
	*q = '\0';
      }

      drawdate = 1;
      break;

    case 'f':
      fontname = strdup(optarg);
      drawdate = 1;
      break;

    case 'R':
      runtime = atol(optarg);
      break;

    case 'S':
      sleeptime = atol(optarg);
      break;

    case 'x':
      nbproc_x = atol(optarg);
      break;

    case 'y':
      nbproc_y = atol(optarg);
      break;

    case 'B':
      bdrsize = atol(optarg);
      break;

    case 'u':
      unmanaged = 1;
      break;

    case '?':
      usage();
      exit(0);
      break;
    }
  }

  if (nbproc_x > 1 || nbproc_y > 1) {
      /* fork mode required */
      pid = getpid();
      for (proc_y = nbproc_y - 1; proc_y >= 0; proc_y--) {
	  for (proc_x = nbproc_x - 1; proc_x >= 0; proc_x--) {
	      pid = fork();
	      if (pid < 0) {
		  perror("fork()");
		  exit(1);
	      }
	      if (pid == 0) {
		  X += proc_x * (width + bdrsize);
		  Y += proc_y * (height + bdrsize);
		  break;
	      }
	  }
	  if (pid == 0)
	      break;
      }

      /* in parent, and all processes already forked ? */
      if (pid != 0)
	  exit(0);
  }

  signal(SIGUSR1, sig_resync);

  dpy = XOpenDisplay(NULL);
  if (!dpy) {
    fprintf(stderr, "Failed to open default display.\n");
    exit(1);
  }

  XSetErrorHandler(errhandler);

  root = DefaultRootWindow(dpy);

  alloc_color(bg, &black);
  alloc_color(bg2, &green);
  alloc_color(fg, &orange);
  alloc_color(fg2, &blue);

  swa.override_redirect = unmanaged;

  win = XCreateWindow(dpy, root, X, Y, width, height, 0,
  			    CopyFromParent, InputOutput, CopyFromParent,
  			    CWOverrideRedirect, &swa);
  XStoreName(dpy, win, "ocbench");

  XSelectInput(dpy, win, ExposureMask | StructureNotifyMask);
  XMapWindow(dpy, win);
  XMoveWindow(dpy, win, X, Y);

  buffer = XCreatePixmap(dpy, win, width, height, 
			 DefaultDepth(dpy, DefaultScreen(dpy)));


  gc = XCreateGC(dpy, win, 0, &gc_setup);
  XSetArcMode(dpy, gc, ArcPieSlice);

  copygc = XCreateGC(dpy, win, 0, &gc_setup);
  XSetFillStyle(dpy, copygc, FillTiled);

  font = XLoadQueryFont(dpy, fontname);
  if (!font) {
    fprintf(stderr, "Font not found, using fixed.\n");
    font = XLoadQueryFont(dpy, "fixed");
  }
  XSetFont(dpy, gc, font->fid);

  while (1) {
    struct timeval tv;
    XEvent ev;
    if (XPending(dpy)) {
      XNextEvent(dpy, &ev);
      switch(ev.type) {
      case Expose:
	expose((XExposeEvent *)&ev);
	break;
      case ConfigureNotify:
	width = ev.xconfigure.width;
	height = ev.xconfigure.height;
	XFreePixmap(dpy, buffer);
	buffer = XCreatePixmap(dpy, win, width, height, 
			       DefaultDepth(dpy, DefaultScreen(dpy)));
	draw();
	break;
      }
      continue;
    }
    draw();

    /* any runtime ? */
    if (runtime > 0) {
	    struct timeval tv2, tv3;
	    gettimeofday(&tv2, NULL);
	    do {
		    gettimeofday(&tv3, NULL);
	    } while ((1000000+tv3.tv_usec-tv2.tv_usec)%1000000 < runtime);
    }

    tv.tv_usec = sleeptime; tv.tv_sec = 0;
    select(0, NULL, NULL, NULL, &tv);
  }
}
